/* * EMBL-EBI
* MSD Group
* VAMSAS Project
* VAMAV
*  * @author <a href="mailto:pierre@ebi.ac.uk">Pierre MARGUERITE</a>
 * @version  1  
*/
package uk.ac.ebi.msd.vamsas.vamav.documentmanagement.grouping;

import uk.ac.ebi.msd.vamsas.vamav.documentmanagement.grouping.exceptions.UnableToGroupSequencesException;
import uk.ac.ebi.msd.vamsas.vamav.exceptions.IncorrectVamavModuleInitialisation;
import uk.ac.ebi.msd.vamsas.vamav.util.data.DataBuffer;
import uk.ac.ebi.msd.vamsas.vamav.util.logger.VAMAVLogger;
import uk.ac.ebi.msd.vamsas.vamav.util.logger.VAMAVLoggerImpl;
import uk.ac.ebi.msd.vamsas.vamav.util.mapping.MappingManager;
import uk.ac.ebi.msd.vamsas.vamav.util.mapping.exceptions.MappingException;
import uk.ac.ebi.msd.vamsas.vamav.util.mapping.matrixrotation.RotationTranslationMatrix;
import uk.ac.ebi.msd.vamsas.vamav.util.multiplealignment.AlignmentManager;
import uk.ac.ebi.msd.vamsas.vamav.util.multiplealignment.ClustalW_EBIWSClient;
import uk.ac.ebi.msd.vamsas.vamav.util.multiplealignment.MultipleAlignmentIntf;
import uk.ac.ebi.msd.vamsas.vamav.util.multiplealignment.exceptions.AlignmentException;
import uk.ac.ebi.msd.vamsas.vamav.util.pool.DataBufferPool;
import uk.ac.ebi.msd.vamsas.vamav.util.pool.StringBufferPool;
import uk.ac.ebi.msd.vamsas.vamav.util.properties.PropertyManager;
import org.vamsas.objects.core.AnnotationElement;
import org.vamsas.objects.core.SequenceType;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;

/**
* Abstract class for defining Structure grouping methods. 
* 
* Grouping mode should implement these methods
* 
* Providing method definitions and code for atribute file generation
 * @author <a href="mailto:pierre@ebi.ac.uk">Pierre MARGUERITE</a>
 * @version  1  
*/

public abstract class SequenceGrouping  implements uk.ac.ebi.msd.vamsas.vamav.VamaVModule{

	protected final String  pdbidsproperty = "pdbids";
	
	protected  String modulePrefix = "sequence";
	protected final VAMAVLogger logger = new VAMAVLoggerImpl(this.getClass());

	protected  Hashtable <String, ArrayList<SequenceType>> groups = null;
	
	/**
	 * List of PDB structures per group
	 * to allow rotation translation matrix mapping
	 */
	protected  Hashtable <String, ArrayList<String>> structuresPerGroup = null;
	
	protected char gapChar = 'X';
	protected Hashtable <String,Integer> maxSeqLengthsPerGroup = null;
	protected Hashtable <Integer,String> fakeSequencePerLength = null;
	
	protected ArrayList<File> sequenceFile = null;
	protected final String atrFilePrefix = "atr";
	protected final String alignmentFilePrefix = ".aln";
	protected String workingDirectory = "atributeFile";
	//protected String atributeFile = this.workingDirectory+File.separator+"vamsas"+"."+this.atrFilePrefix;
	protected static String enc = null;
	protected static final boolean erase = true;
	protected static final boolean append = false;
	protected Hashtable<String, Hashtable<Integer, StringBuffer>> sequenceBreakPoints = null;
	protected final boolean deleteOnExit = false;
	/**
	 * Mapping manager to retrieve a mapping between ID for example
	 * 
	 */
	protected MappingManager mapping = null;
	
	 protected AlignmentManager alignManager = null;
	
	    
	/**
	 * <p></p>
	 * 
	 */
	 //   private MultipleAlignmentIntf multipleAlignment;

	
	  /**
	   * List of breakpoints for the current alignmentSet
	   */
	protected  ArrayList<AnnotationElement> breakpoints = null;
	  /**
	   * Saves adding order of sequences 
	  */
	protected  ArrayList <String> groupOrder = null;
	
	/**
	 *  
	 * Constructor of the  SequenceGrouping class, given a mapping manager 
	 *
	 * @param _mappingManager the mapping manager to use for mapping 
	 */
	  public SequenceGrouping( MappingManager _mappingManager)
	    	{
	    		this.mapping = _mappingManager;
	    		
	    	}
	
	/**
	 * Inits the Sequence Grouping sub module, with a list of configuration properties.
	 */
	public void init (Properties moduleProperties) throws IncorrectVamavModuleInitialisation
		{
			PropertyManager propManager = new PropertyManager();
			this.alignManager = new  AlignmentManager();
			try
				{
					Properties props = propManager.getModuleProperties(moduleProperties, this.alignManager.getModulePrefix());
					this.alignManager.init(props);
					props = null;
				}
			catch (IncorrectVamavModuleInitialisation ivmie)
				{
					logger.appliError("Unable to init Alignment Manager.");
					throw new IncorrectVamavModuleInitialisation(ivmie);
				}
			propManager = null;
			if (moduleProperties == null || moduleProperties.size()<1)
				{
					this.logger.appliInfo("SequenceGrouping - No properties provided. Uses default.");
					return;
				}
			String gap = moduleProperties.getProperty("gapcharacter", String.valueOf(this.gapChar));
			if( gap !=null && gap.length()<1) 
				this.gapChar = gap.charAt(0);
		
		}
	 

    
    
  
    
    /**
	 * Retrieves the  mapping attribut
	 *
	 * @return Returns the mapping.
	 */
	protected MappingManager getMapping()
		{
			return this.mapping;
		}

	/**
	 *  retrieves the MultipleAlignmentIntf associated to the current object,
	 *  to execute multiple alignment
	 * @return the MultipleAlignmentIntf object associated to the current object.
	 */
 /*   private  MultipleAlignmentIntf getMultipleAlignment() {
        return this.multipleAlignment;
    }*/
    
    /**
     * Set the MultipleAlignmentIntf for the current object
     * @param multipleAlignmentIntf the MultipleAlignmentIntf object ot associated to the object.
     */
   /* private void setMultipleAlignment(MultipleAlignmentIntf _multipleAlignment) 
    	{
    	    this.multipleAlignment = _multipleAlignment;
    	}*/
    
    /**
     * Add a sequence for grouping.
     * 
     * @param sequence chain sequence to add
     * @param name name of the sequence to add
     * @return a list of parameters to call the AstexViewer
     * @throws UnableToGroupSequencesException if an error occurs during the adding.
     */
    public abstract Hashtable<String, Object>  addSequence(String sequence, String name) throws UnableToGroupSequencesException;
    
    /**
     * Add a sequence for grouping, with an associated accessionID and its source.
     * 
     * @param sequence chain sequence to add for grouping
     * @param name name of the sequence to add
     * @param accessionID accessionId associated to the sequence
     * @param source source of the accessionID
     * @return list of parameters to call the AstexViewer, if needed.  Null, otherwise.
     * @throws UnableToGroupSequencesException if an error occurs during the adding.
     */
    public abstract Hashtable<String, Object> addSequence(String sequence, String name, String accessionID, String source) throws UnableToGroupSequencesException;
    
    /**
     * Registers a PDB ID as belonging to a group
     * @param groupName name of the group to add the structure in it.
     * @param pdbID Id of the PDB structure to add
     */
    protected void addStructureToGroup(String groupName, String pdbID)
    	{
    		if (this.structuresPerGroup == null)
    			{
    				this.structuresPerGroup = new Hashtable<String,ArrayList<String>>();	
    			}
    		ArrayList<String> structs= this.structuresPerGroup.get(groupName);
    		if (structs == null)
    			{//create new list of the group
    				structs = new ArrayList<String> ();
    			}
    		//add structure
    		if (!structs.contains(pdbID))
    			{
    				structs.add(pdbID);
    			   		logger.sessionDebug("Adding structure"+pdbID+" to group "+groupName);
    				//	readd list to be sure to have the right list in the global list
    				this.structuresPerGroup.put(groupName, structs);
    			}
    	}
    /**
	   * Add a sequence to a group
	   * @param groupName name of the group to add the sequence into
	   * @param sequence sequence to add
	   * @param sequenceName name of the sequence to add
	   */
    protected void addSequenceToGroup(String groupName, String sequence, String sequenceName)
    	{
    		if (groupName == null || groupName.length()<1) 
    			{
    				this.logger.sessionError("There is no group name. Can not add the sequence to the group.");
    				return;
    			}
    		if (sequence == null || sequence.length()<1) 
    			{
    				this.logger.sessionError("There is no sequence. Can not add the sequence to the group.");
    				return;
    		    }
    		    
    		this.logger.sessionInfo("Adding "+sequenceName+" sequence to "+groupName+" group.");
    		ArrayList<SequenceType> elements = getGroups().get(groupName);
    		if (elements == null) 
    			{
    				elements = new ArrayList<SequenceType>();
    				this.logger.sessionInfo("No previous list, create one");
    			}
    		if (sequence ==null)
    			this.logger.sessionError("No sequence provided for "+sequenceName+".");
    		else
    			{
    				if (sequence.length()>0)
    					{
    						Integer maxLength = getMaxSeqLengthsPerGroup().get(groupName);			
    						if(maxLength!=null&& maxLength.intValue()>0 )
    							{
    								int maxSeqLength = maxLength.intValue();
    								int seqLen = sequence.length();
    								if (maxSeqLength < seqLen)
    									{
    										maxSeqLength = seqLen;
    										maxLength = Integer.valueOf(maxSeqLength);
    										getMaxSeqLengthsPerGroup().put(groupName, maxLength);
    									}
    		    					}
    		    				else
    		    					{
    		    						int seqLen = sequence.length();
    		    						maxLength = Integer.valueOf(seqLen);
    		    						getMaxSeqLengthsPerGroup().put(groupName, maxLength);
    		    					}
    					}
    				SequenceType seq = new SequenceType() ;
    				seq.setName(sequenceName);
    				seq.setSequence(sequence);
    				if( !elements.contains(seq))
    					{
    						if (!this.logger.isProduction())
    							{
    								this.logger.sessionDebug("sequences size before ");
    								this.logger.sessionDebug(""+elements.size());
    							}
    						elements.add(seq);
    					}
    				//else
    					//seq = null;
    				if( !this.logger.isProduction())
    					{
    						this.logger.sessionDebug("sequences size after ");
    						this.logger.sessionDebug("" +elements.size());
    						this.logger.sessionDebug("groups size before ");
    						this.logger.sessionDebug(""+getGroups().size());
    					}
    				getGroups().put(groupName, elements);
    				if( !this.logger.isProduction())
    					{
    						this.logger.sessionDebug("groups size after ");
    						this.logger.sessionDebug(""+getGroups().size());
    					}
    				addGroupToOrder(groupName);
    			}
    	}

	/**
	 * Retrieves the  groups attribut
	 *
	 * @return Returns the groups.
	 */
    protected  Hashtable<String, ArrayList<SequenceType>> getGroups()
		{
			if (this.groups == null)
				this.groups = new  Hashtable<String, ArrayList<SequenceType>> ();
			return this.groups;
		}

	/**
	 * Retrieves the  fakeSequencePerLength attribut
	 *
	 * @return Returns the fakeSequencePerLength.
	 */
	public Hashtable<Integer, String> getFakeSequencePerLength()
		{
			if (this.fakeSequencePerLength == null)
				{
					this.fakeSequencePerLength = new Hashtable<Integer, String>();
				}
			return this.fakeSequencePerLength;
		}

	/**
	 * Retrieves the  maxSeqLengthsPerGroup attribut
	 *
	 * @return Returns the maxSeqLengthsPerGroup.
	 */
	public  Hashtable<String, Integer> getMaxSeqLengthsPerGroup()
		{
			if( this.maxSeqLengthsPerGroup == null)
				this.maxSeqLengthsPerGroup = new Hashtable<String, Integer>();
			return this.maxSeqLengthsPerGroup;
		}
    
	  /**
	   * Retrieves the  sequenceOrder attribut
	   *
	   * @return Returns the sequenceOrder.
	   */
	  protected  ArrayList<String> getGroupOrder() {
	    if( this.groupOrder == null)
	    	this.groupOrder = new ArrayList <String>();
	    return this.groupOrder;
	  }
	  
	  protected  void addGroupToOrder(String groupName)
	  {
	    if (!(getGroupOrder().contains(groupName)))
	      getGroupOrder().add(groupName);
	  }
	  
	  /**
	   * Clear the sequence order List
	   */
	  public  void groupOrderClear() {
	    if( this.groupOrder != null)
	    	this.groupOrder = null;
	  }

	/**
	 * Retrieves the  modulePrefix attribut
	 *
	 * @return Returns the modulePrefix.
	 */
	public String getModulePrefix()
		{
			return this.modulePrefix;
		}

	 /**
	   * Add a breakpoint to the grouping
	   * @param element annotation element associated to a breakpoint.
	   */
	  public void  addBreakpoint(AnnotationElement element)
		  {
			  if (element == null) 
			    {
			    	this.logger.sessionError("Sequence Grouping  addBreakPoint - There is no annotation element. Can not add the breakpoint to groups");
			    	return;
			    }
			  this.logger.sessionDebug("Adding breakpoint");
			  getBreakpoints().add(element);
		  }
	  
	  /**
	  * Retrieves the  breakpoints attribut
	  *
	  * @return Returns the breakpoints.
	  */
	  protected ArrayList<AnnotationElement> getBreakpoints()
	 	{
	 		if (this.breakpoints == null)
	 			this.breakpoints= new ArrayList<AnnotationElement>();
	 		return this.breakpoints;
	 	}
	 
	 /**
	  * Create a fake sequence containing provided breakpoints.
	  * @param breaks list of breakpoints
	  * @param groupName name of the group the fake sequence will be associated.
	  * @return the created fake sequence corresponding to the list of breakpoint
	  */
	  protected String  createFakeSequence(String groupName, ArrayList<AnnotationElement> breaks)
	 	{
	 		
	 		if(breaks == null || breaks.size()<1) return null;
	 		this.logger.sessionInfo("Creating BreakPoint sequence for "+groupName );
	 		this.logger.sessionDebug("break points: "+breaks.size());
	 		String fakeSequence = null;
	 		Integer len = getMaxSeqLengthsPerGroup().get(groupName);
	 		if (len != null)
	 			{
	 				//retrieves store fake sequence:
	 				fakeSequence = getFakeSequencePerLength().get(len);
	 				if( fakeSequence == null || fakeSequence.length()<1)
	 					{
	 						Hashtable<Integer, AnnotationElement> positions  = new Hashtable<Integer, AnnotationElement> (breaks.size());
	 						StringBuffer sequence = StringBufferPool.getString(); 
	 						for (int breakNumber = breaks.size()-1; breakNumber >=0; breakNumber --)
	 							{
	 								AnnotationElement bp = breaks.get(breakNumber);
	 								if (bp == null) continue;
	 								int pos = bp.getPosition()-1;//position in the sequence start at 1
	 								if (pos >=0)
	 									{
	 										positions.put(new Integer(pos), bp);
	 										//System.out.println("adding position "+pos);
	 									}
	 							}
	 						
	 						int maxSeqLength = len.intValue();
	 						int size = maxSeqLength -1;
	 						for (int i =  0; i< maxSeqLength; i++ )
	 							{
	 								int reversePos =  i;
	 								//System.out.println("reverse pos "+reversePos);
	 								Integer p = new Integer(reversePos);
	 								if (positions.containsKey(p))
	 									{
	 										AnnotationElement bp = positions.get(p);
	 										sequence.append(bp.getDisplayCharacter());
	 									}
	 								else
	 									{
	 										sequence.append(this.gapChar);
	 									//System.out.println("no breakpoint found at pos "+reversePos);
	 									}
	 							}
	 				positions = null;
	 				fakeSequence = sequence.toString();
	 				StringBufferPool.returnString(sequence);
	 				getFakeSequencePerLength().put(len, fakeSequence);
	 			}
	 			}
	 		else
	 			{
	 				this.logger.sessionDebug("no len");
	 			}
	 		this.logger.sessionDebug("fake sequence "+fakeSequence);
	 		return fakeSequence;
	 	}


	 /**
	  * Writes created groups into an atribute file for tha AstexViewer
	  * 
	  * @throws UnableToGroupSequencesException, if unable to group added sequences
	  *  @return the atribute file corresponding to the grouping of added sequences
	  *
	  */ 
	 public File storeGroups() throws  UnableToGroupSequencesException
		 {
			 int size = this.getGroups().size();
			 if (size<1)
				 {
					 this.logger.sessionWarn("Sequence Grouping - There is no computed group. Can not store groups.");
					// return null;
					 throw new UnableToGroupSequencesException();
				 }
			
			 //test if there is at least a sequence in each group. 
			 boolean correct = true;
			 for (Enumeration<String> _groups = this.getGroups().keys(); _groups.hasMoreElements() && correct; )
				 {
					 String groupName = _groups.nextElement();
					 ArrayList<SequenceType> sequences = this.getGroups().get(groupName);
					 this.logger.sessionDebug("Group "+groupName);
					 if (sequences == null || sequences.size()<1)
						 {
							 correct = false;
						 }
					 else
						 {
							 this.logger.sessionDebug("group size "+sequences.size());
						 }
				 }
			 if (!correct)
				 {
					 this.logger.sessionWarn("Sequence Grouping - There is no computed group. Can not store groups.");
					 //return null;
					 throw new UnableToGroupSequencesException();
				 }
			 if (this.sequenceFile == null)
				 this.sequenceFile = new ArrayList<File>();
			 else 
				 this.sequenceFile.clear();
			 File atrFile = null;
			 try
				 {
					 this.workingDirectory  = this.getOutputDirectory();
					 String atributeFile = this.workingDirectory+File.separator+"vamav-"+this.getSessionID()+"."+this.atrFilePrefix;
					 logger.sessionInfo("Creating atribute file: "+atributeFile);
					 //writing Atribute file
					 if( atributeFile == null)
						 {
							 throw new UnableToGroupSequencesException("No atribute file path provided.");
						 }
					 if (!atributeFile.endsWith(this.atrFilePrefix))
						 {
							 atributeFile = atributeFile.concat("."+this.atrFilePrefix);
						 }
					 atrFile = new File (atributeFile);
					 //atrFile.mkdirs();
					 this.logger.sessionInfo("SequenceGrouping - Writing Atribute File.");
					 PrintWriter out = openFile(atrFile);
					 if( out == null)
						 throw new UnableToGroupSequencesException("Unable to open file to write. "+atrFile.getAbsolutePath());;
						 //return null;
					 DataBuffer line = DataBufferPool.getString();
	      
					 //write header
					 line.setString("Atribute file Version 1.0");
					 writeLine(line, out);
					 line.setString("");
					 writeLine(line, out);
					 /*int groupNumber = 0;
					  for (Enumeration keys = getGroups().keys(); keys.hasMoreElements();)
					  {*/
					
					 this.alignManager.setSession(this.getSessionID());
					 for (int groupNumber = 0; groupNumber < getGroupOrder().size() ;groupNumber++ )
						 {
							 
							 
							 String groupName = getGroupOrder().get(groupNumber);
							 getGroups();
							 //exporting fasta file
							 ArrayList<SequenceType> sequences = getGroups().get(groupName);
							 
							 //generate alignment file name
							// MultipleAlignmentIntf 
							// AlignmentManager clustal =  new  AlignmentManager();//ClustalW_EBIWSClient(this.getSessionID());
							 //clustal.setSession(this.getSessionID());
							 String extension = "";
							 String alignmentfileExtension = alignManager.getMultipleAligmnentFileExtension();
							 if (alignmentfileExtension != null && alignmentfileExtension.length()>0)
								if (alignmentfileExtension.startsWith("."))
									extension = alignmentfileExtension;
								else
									extension = ".".concat(alignmentfileExtension);
							 
							 String alignmentFileName = groupName + extension;
							 if( sequences != null && sequences.size()>0)
								 {
									 //(groupName, "atributeFile");
									 this.logger.sessionDebug("Preparing multiple alignment "+groupName);
									 if(getBreakpoints()!=null && getBreakpoints().size()>0)
										 {
											 alignManager.HasBreakPoint();
											 String sequence = createFakeSequence(groupName,getBreakpoints() );
											 if (sequence !=null)
												 {
													 String sequenceName = groupName+"Breakpoints_";
													 alignManager.addSequence(sequenceName, this.prepareAddSequence(sequence,sequenceName));
												 }
										 }
									 else
										 {
											 alignManager.HasNoBreakPoint();
										 }
									 for (int seqNumber = sequences.size() -1; seqNumber>=0; seqNumber--)
										 {
											 SequenceType seq = sequences.get(seqNumber);
											 if (seq == null) 
												 {
													 this.logger.sessionWarn("No sequence to add");
													 continue;
												 }
											 this.logger.sessionDebug("Add sequence for multiple alignment "+seq.getName() + " "+seq.getSequence());
											 alignManager.addSequence(seq.getName(), this.prepareAddSequence(seq.getSequence(), seq.getName()));
										 }
									String alignment = alignManager.execMultipleAlignment();
									if (alignment!= null)
										{
											File alignmentFile = this.unModifieSequences(alignment, groupName,this.workingDirectory + File.separator+alignmentFileName);
											if (alignmentFile != null)
												this.getSequenceFiles().add(alignmentFile);
										}
									else
										{
											logger.sessionDebug("No alignment retrieved.");
										}
									
			        //	add group
									 ArrayList<String> structs = null;
									 DataBuffer ids = DataBufferPool.getString();
									
									 if( this.structuresPerGroup != null && this.structuresPerGroup.size()>0)
										 {
											 structs = this.structuresPerGroup.get(groupName);
											 
											 if (structs == null || structs.size()<1)
												 {
													 ids.append("groupName");
													 
													 logger.sessionInfo("no pdb structure for the group "+groupName+". No matrix added");
													 line.setString("  Molecule "+groupName+" "+groupName+"  {\n");
													// writeLine(line, out);
													 line.append("  }\n");
													 //writeLine(line, out);
												 }
											 else
												 {
													 String primaryID = structs.remove(structs.size()-1);
													 line.setString("  Molecule "+primaryID+" {\n");
													// writeLine(line, out);
													 line.append("  }\n"); 
													 //writeLine(line, out);
													 ids.append(primaryID);
													 for (int structNumber = structs.size()-1; structNumber>=0; structNumber--)
														 {
															 String id = structs.get(structNumber);
															 line.append("  Molecule "+id+" {\n");
															 //writeLine(line, out);
															 ids.append(" ".concat(id));
															 try
																 {
																	 Collection<RotationTranslationMatrix> matrixes = this.mapping.getRotationTranslationMatrix(primaryID, id);
																	 if (matrixes != null)
																		 {
																		 	for (RotationTranslationMatrix matrix : matrixes)
																			 {
																			 	if( matrix != null)
																		 			{
																			 			boolean isChain = false;
																			 			String chainID = matrix.getPrimaryPDBIDChain();
																			 			if(chainID!= null  &&chainID.length()>0 )
																					 		{
																						 	  isChain = true;
																					 		  line.append("    Chain "+chainID+ " {\n");
																				 			}
																			 			if( matrix != null)
																				 			{
																					 			line.append("    Matrix ");
																					 			 line.append(BigDecimal.valueOf(matrix.get(0, 0).doubleValue())+ " "+BigDecimal.valueOf(matrix.get(0,1).doubleValue())+ " "+BigDecimal.valueOf(matrix.get(0,2).doubleValue())+ " "+
																							 	BigDecimal.valueOf(matrix.get(1,0).doubleValue())+ " "+BigDecimal.valueOf(matrix.get(1, 1).doubleValue())+ " "+BigDecimal.valueOf(matrix.get(1, 2).doubleValue())+ " "+
																							 	BigDecimal.valueOf(matrix.get(2,0).doubleValue())+ " "+BigDecimal.valueOf(matrix.get(2, 1).doubleValue())+ " "+BigDecimal.valueOf(matrix.get(2,2).doubleValue())+ " "
																							 	+ BigDecimal.valueOf(matrix.get(0,3).doubleValue()) + " " + BigDecimal.valueOf(matrix.get(1, 3).doubleValue()) + " " + BigDecimal.valueOf(matrix.get(2, 3).doubleValue()) );
																			 //writeLine(line, out);
																					 			line.append("\n");
																				 			}
																			 			if( isChain)
																				 			{
																					 			line.append("}\n"); 
																				 			}
																		 			}
																			 	}
																		 }
																 }
															 catch (MappingException mexcep)
																 {
																	 logger.sessionError("Error during RTM mapping.",mexcep);
																 }
															 line.append("  }\n"); 
															 //writeLine(line, out);
														 }
													 
												 }
											 
										 }
									 DataBuffer groupLine = DataBufferPool.getString();
									groupLine.setString("Group \""+groupName+"\" "+ids+" {");
									// groupLine.setString("Group \""+groupName+"\" {");
									 writeLine(groupLine, out);
									 DataBufferPool.returnString(groupLine);
									 DataBufferPool.returnString(ids);
									 line.append("  Sequence \""+alignmentFileName+"\" \n");//\""+groupName+".aln"+"\" ");
									 //writeLine(line, out);
									 line.append("  Text \"This is a group "+groupName+"\"\n");
									 //writeLine(line, out);
									 writeLine(line, out);
			            //Group "dfr" 7dfr 8dfr {
									 
									 line.setString("}\n");
									 writeLine(line, out);
									
									
									
								 }
							 else 
								 this.logger.sessionWarn("No sequence to add to group "+ groupName);
						 }
					 line.setString("}\n");
					 writeLine(line, out);
					 closeFile(out);
					 alignManager = null;	
				 } 
			 catch (AlignmentException ae )
				 {
					 atrFile = null;
					 throw new UnableToGroupSequencesException(); 
				 }
			 catch (IOException ioe)
				 {
					 this.logger.sessionError("Error during Atribute file writing "+atrFile+". ",ioe);
					 //ioe.printStackTrace(System.out);
					 atrFile = null;
				 }
			 return atrFile;
	  }
	  
	 /**
	  * Readds breakpoints after the multiple alignment ( as  display character for breakpoint, are not supported by multiple alignment tools
	  * 
	  * @param alignment alignment String 
	  * @param alignmentName name of the alignment
	  * @return the file containing the alignment with breakpoints
	  */
	 protected File unModifieSequences (String alignment, String alignmentName, String masterFileName)
		 {
			 if (alignment == null) 
				 {
					 this.logger.sessionError("UnModifiedSequences - No aligment found. Can not unmodified sequences.");
					 return null;
				 }
			// logger.Applidebug("Readded breakpoint to sequences "+alignment);
			
			 
			// String masterFileName = this.workingDirectory + File.separator+alignmentName+ alignmentfileExtension;
		      // WORK_ROOT + "/" +sessionId+ "/master_" + i + ".seq";
		     //check workingDirectory exists, can be read and is a directory
			 File workingDir = new File (this.workingDirectory);
			 if(!(workingDir !=null && workingDir.exists()))
		   	      {
		   	    	  if (!workingDir.mkdir())
		        	   {
		        		   this.logger.sessionError( "%%% WARNING: couldn't open master "
		        				   + "sequence file for writing. Working directory does not exist. Can not be created. "+this.workingDirectory );
		        		   return null;
		        	   }	
		   	      }
			 if(! (workingDir.isDirectory()&&  workingDir.canRead()))
				 {
		    	     logger.sessionError( "%%% WARNING: couldn't open master "
		    	    		 + "sequence file for writing. Working directory does not exist. " );
		    	     return null;
		       }
			File masterFile = new File(masterFileName );
			 try
				 {
					 Writer masterWriter = new FileWriter( masterFile );
					 PrintWriter master = new PrintWriter( masterWriter );
					 if(this.deleteOnExit) masterFile.deleteOnExit();
		         
		  	       //write alignment
					 boolean endAlignment = false;
					 int len = 0;
					 DataBuffer brkSequence = DataBufferPool.getString();
					 logger.sessionDebug("Retrieving alignment");
					 String[] lines = alignment.split("\n");
					 if (lines == null || lines.length<1)
						 {
							 logger.sessionError("No line retrieving - something wrong.");
							 return null;
						 }
					 logger.sessionDebug(lines.length+" lines found.");
					 for ( int lineNumber = 0; lineNumber<lines.length; lineNumber++ )
						 {
							 String line = lines[lineNumber];
							 	if (line == null)continue;
							 	//modified breakpoint sequence as incorrect characters are erased (they have been replaced by X)
							 	if(line.trim().length()<=0 ||line.trim().equals(" ") ) {master.println(line); continue;}
							 	if (line.startsWith("Clustal") || line.startsWith("CLUSTAL")) 
							 		{
							 			master.println(line);
							 			continue;
							 		}
							 	//System.out.println(line);
							 	int pos = line.indexOf(" ");
							 	String name = line.substring(0,pos);
							 	name = name.trim().toLowerCase();
							 	Hashtable<Integer, StringBuffer> bkpts = this.getSequenceBreakPoints().get(name);
							 	if (bkpts == null)//no breakpoint associated - write line
							 		{
							 			logger.sessionDebug("No breakpoint export line "+line);
							 			master.println(line);
			    			/*System.out.println("no break point");*/
							 		}
							 	else
							 		{//modified at positions
							 			//get sequence
							 			/*System.out.println(line);*/
							 			logger.sessionDebug("No breakpoint found ");
							 			String sequence = line.substring(pos+1,line.length()).trim();
							 			/*System.out.println("sequence: "+sequence);*/
							 			int seqPos = line.indexOf(sequence);
							 			/*System.out.println("break point "+bkpts.size());*/
			    		
							 			char[] chars = sequence.toCharArray();
							 			int currentResidentNumber = 0;
							 			for (int charNumber = 0; charNumber< chars.length;charNumber++ )
							 				{
							 					char currentChar = chars[charNumber];
							 					if (currentChar == '-') 
							 						{
							 						//System.out.print(" gap ");
							 						continue;
							 						}
							 					//check if breakpoint
							 					/*System.out.print(" "+currentResidentNumber+" ");*/
							 					Integer resN = Integer.valueOf(currentResidentNumber++);
							 					StringBuffer displayCharacter = bkpts.get(resN);
							 					len++;
			    					//System.out.print("|"+(currentResidentNumber-1)+"-"+len+"|");
							 					if (displayCharacter == null ||displayCharacter.length()<1 )
							 						{
							 							continue;
							 						}
			    						//breakpoint found
							 					chars[charNumber] = displayCharacter.charAt(0);
			    						/*System.out.print(displayCharacter.charAt(0));*/
							 					bkpts.remove(resN);resN=null;
							 				}
							 			int resCount = currentResidentNumber;
							 			sequence = String.valueOf(chars);
			    				/*System.out.println("new seq: "+sequence);*/
							 			if (brkSequence.length()<=0) brkSequence.setString(sequence);
							 			else brkSequence.append(sequence);
							 			line = line.substring(0,seqPos).concat(sequence);
							 			master.println(line);
							 			Hashtable<Integer, StringBuffer> newBkpts = new Hashtable<Integer, StringBuffer> (bkpts.size());
					    
							 			for (Enumeration positions = bkpts.keys(); positions.hasMoreElements();)
							 				{
							 					Integer position = (Integer) positions.nextElement();
							 					if (position == null|| position.intValue()<0) continue;
							 					StringBuffer character = bkpts.get(position);
							 					if( character == null || character.length()!=1)
							 						continue;
			    							/*System.out.println("previous breakpoint position "+position.intValue());*/
			    								bkpts.remove(position);	
			    						//position = null;
			    								position = Integer.valueOf(position.intValue() - (resCount));
			    							/*System.out.println("new breakpoint position "+position.intValue());*/
			    							//use temporary table to avoid deadlock
			    								newBkpts.put(position, character);
							 				}
			    		/*for (Enumeration positions = bkpts.keys(); positions.hasMoreElements();)
			    			{
			    				Integer position = (Integer) positions.nextElement();
			    				if (position == null|| position.intValue()<0) continue;
			    				StringBuffer character = bkpts.get(position);
			    				if( character == null || character.length()!=1)
			    					continue;
			    				if( position.intValue()<sequence.length()-1 && position.intValue()<=0)
			    					{
			    						StringBuffer newLine = new StringBuffer(line);
			    						newLine.setCharAt(position.intValue(), character.charAt(0));
			    						line = newLine.toString();
			    						newLine = null;
			    						bkpts.remove(position);
			    					}
			    				else
			    					{
			    						position = Integer.valueOf(position.intValue() - (sequence.length()));
			    						//use temporary table to avoid deadlock
			    						tempTable.put(position, character);
			    					}
			    			}
			    		for (Enumeration positions = tempTable.keys(); positions.hasMoreElements();)
			    			{
			    				Integer position = (Integer) positions.nextElement();
			    				if (position == null|| position.intValue()<0) continue;
			    				StringBuffer character = bkpts.get(position);
			    				if( character == null || character.length()!=1)
			    					continue;
			    				bkpts.put(position, character);
			    			}*/
			    		/*System.out.println("New breakpoints "+bkpts.size());
			    		System.out.println("New breakpoints "+newBkpts.size());*/
			    		//remove previous breakpoint list
							 			this.getSequenceBreakPoints().remove(name);
							 			bkpts = null;
							 			if (newBkpts != null && newBkpts.size()>0)
							 				this.getSequenceBreakPoints().put(name, newBkpts);		
							 		}
						 }
				       
			         //release unused object
					 	this.clearSequenceBreakPoints();
					 	//StringBufferPool.returnString(this.sequences);
					 	alignment = null;
					 	master.close();
					 	master = null;
					 	alignmentName = null;
					 	// System.out.println("new bk seq "+brkSequence);
					 	DataBufferPool.returnString(brkSequence);
				 }
			 catch( IOException ioe ) 
			       {
			        	logger.sessionError( "%%% WARNING: couldn't open master "+ "sequence file for writing: " + ioe );
			   //	      ioe.printStackTrace(System.out);
			        	masterFile = null;
			       }
			 return masterFile;
			}
	 
	 /**
	  * As multiple alignment tools do not support breakpoint display characters, they are removed from the sequence, replaced by 'X' character 
	  * and stored for readding after the multiple alignment (see unModifieSequences method) 
	  * @param sequence sequence to prepare for alignment
	  * @param name name of the sequence to prepare
	  * @return the sequence without breakpoint display character
	  */
		 
	 protected String prepareAddSequence(String sequence, String name)
		 {
//			save position for breakpoint sequences
				//modified non expected character (	A R D N B C E Z Q G H I L K M F P S T W Y V X)
				//t	o the same/unique one ->"|"
				//
				/*sequences.setLength(0);
				sequences.trimToSize();
				sequences.append(sequence);*/
				//System.out.println("current seq "+sequences);
				String temp = sequence.replaceAll("[^ARDNBCEZQGHILKMFPSTWYVX-]", "|");
				StringBuffer s = StringBufferPool.getString();
				s.setLength(0);
				s.trimToSize();
				s.append(temp);
				int pos = s.indexOf("|");
				//System.out.println("tem p seq "+temp);
				Hashtable<Integer, StringBuffer> breakpointPositions = null;
				while (pos>0)
					{
						logger.sessionInfo("new breakpoint "+pos);
						if (breakpointPositions == null) breakpointPositions = new Hashtable<Integer, StringBuffer>();
						breakpointPositions.put(Integer.valueOf(pos), new StringBuffer(sequence.subSequence(pos, pos+1)));
						s.setCharAt(pos, 'X');
						pos = s.indexOf("|");	
					}
				if(breakpointPositions!=null )
					{//a breakpoint has been found at least
						this.getSequenceBreakPoints().put(name.trim().toLowerCase(), breakpointPositions);
					}
				String seq = s.toString();
				StringBufferPool.returnString(s);
				return seq;
		 }
	 
	  /**
	   * Retrieves the  sequenceFile attribut
	   *
	   * @return Returns the sequenceFile.
	   */
	  public ArrayList<File> getSequenceFiles() {
		  if( this.sequenceFile == null)
			  this.sequenceFile = new  ArrayList<File> ();
	    return this.sequenceFile;
	  }


	  /**
	   * Opens a file for writing
	   *
	   *@param file file to open.
	   *@return a printwriter to writer into it
	   * @exception IOException if an error occurs
	   */
	  protected PrintWriter openFile(File file) throws IOException 
	  	{
	  		PrintWriter out = null;
	  		if (file == null)
	  			{ 
	   	     //Application log(no such file)
	  				throw new IOException("Unknown file");
	  			}
	      
	      //      check if the current file exists
	    file = new File(file.getAbsolutePath());
	    logger.sessionDebug("atribute file "+file.getAbsolutePath());
	    if (!file.exists() )
	        {
	        	  File directory = new File(file.getParent());
	        	  logger.sessionDebug("parent directory "+ directory.getParent());
	              // but first, see if it exists already. If so, we don't need
	              // to run the groups calculation again, so just return and
	              // we're done
	    			   if( directory.exists() ) {
	    				  //
	    			   } else {
	    				   if( ! directory.mkdirs() ) {
	    					  logger.sessionError( "%%% ERROR: Grouping; couldn't create "
	    							   + "directory for grouping /atribute file "  );
	    				   }
	    			   }
	        if (!file.createNewFile()) 
	          {
	            throw new IOException("Can't create file: "
	                + file.getAbsolutePath());
	          }
	       }
	      // the file will be overwritten, if the file already exists (but it should be indicated in the log)
	      
	      if (file.exists() && !erase)
	        {
	          throw new IOException("The file already exists: "+ file.getAbsolutePath()+ "  Shall we overwrite?");
	        }
	      	

	      if (!file.canWrite() )
	        {
	          throw new IOException("Can not write in file: "
	              + file.getAbsolutePath());
	        }

	      // check possible problem with the reader
	      //  The file is already opened. So, we close the file, before reopening it.
	      if (out != null)
	        {
	          closeFile(out);
	          logger.sessionWarn("The file "+ file.getName()+" has already been opened.");
	        }

	      if (enc != null)
	        {
	          logger.sessionDebug("Opening File: "+file.getName());
	          logger.sessionDebug("set output with encoding : "+enc+"  "+file.getAbsolutePath());
	          FileOutputStream fos = new FileOutputStream(file.getAbsolutePath(), append);
	          try
	        	  { 
	        		  OutputStreamWriter ow = new OutputStreamWriter(fos, enc);
	        		  out = new PrintWriter(ow,true);
	        		  logger.sessionDebug("Encoding "+ow.getEncoding()+" support "+java.nio.charset.Charset.isSupported(enc));
	         	 }
	          catch (java.io.UnsupportedEncodingException ue)
	        	  {
	        		  logger.sessionError("unsupported encoding ",ue);
	        	  }
	    	   }
	      else 
	    	  {
	    		 
	    		  logger.sessionDebug("Opening File: "+file.getName());
	    		  FileWriter fw = new FileWriter(file.getAbsolutePath(), append);
	    		  out = new PrintWriter(new BufferedWriter(fw),true);
	    	  }
	      return out;
	  }
	  
	  
	  /**
	   * Closes a data text file.
	   * @param out writer of the file to close.
	   * 
	   */
	  protected  void closeFile(PrintWriter out)
		  {
			  if (out != null) 
				  {
					  out.flush();
					  out.close();
					  out = null;
				  }
		  }
	  
	  
	  /**
	   * <p>
	   * Retrieves a line from a data file  
	   * Only when termination requirements have been met, the method should return a
	   * <code>String</code> object.
	   * </p>
	   * @param line containing data items in one String
	   * @param out PrintWriter of the file to write line into. 
	   */
	  protected void writeLine(DataBuffer line, PrintWriter out )
	  {
	    /*System.out.println("Writeline ");*/
	    if (line != null)
	      {
	        if (line.length()  == 0) 
	          {
	            /*System.out.println("line empty");*/
	            out.println("");
	          }
	        else
	          {
	            out.println(line);
	           /* System.out.println("Writeline :" + line);*/
	          }
	      }
	  }
	  

		/**
		 * Retrieves the  sequenceBreakPoints attribut
		 *
		 * @return Returns the sequenceBreakPoints.
		 */
	  protected Hashtable<String, Hashtable<Integer, StringBuffer>> getSequenceBreakPoints()
			{
				if(this.sequenceBreakPoints == null)
					{
						this.sequenceBreakPoints = new Hashtable<String, Hashtable<Integer, StringBuffer>>();
					}
					
				return this.sequenceBreakPoints;
			}
		
		/**
		 * Clears breakpoint lisr
		 *
		 */
	  protected void clearSequenceBreakPoints()
			{
				if(this.sequenceBreakPoints != null)
					{
						this.sequenceBreakPoints.clear();
						this.sequenceBreakPoints = null;
					}
			}

		/**
		 * Retrieves the  pdbidsproperty attribut
		 *
		 * @return Returns the pdbidsproperty.
		 */
		public String getPdbidsproperty()
			{
				return this.pdbidsproperty;
			}
		
		/**
		 * Retrieves the working directory for output file.
		 * @return the path to the output/working directory
		 */
		public abstract String getOutputDirectory();
		
		/**
		 * retrieves current sessionID
		 * @return the ID of the current session for the document grouping.
		 */
		public abstract String getSessionID();
 }
